from typing import List, Union, Tuple, Optional
import svg
from .exon import Exon, Region
import math


def shift(
    points: List[Union[float, int]], x_offset: float, y_offset: float
) -> List[Union[float, int]]:
    """Shift the x- and y position by the supplied offsets"""
    return [x + y_offset if i % 2 else x + x_offset for i, x in enumerate(points)]


def draw_exons(
    exons: List[Exon],
    reverse: bool,
    scale: int = 1,
    max_width: float = 1024,
    height: Optional[float] = None,
    non_coding: bool = False,
    gap_size: Optional[int] = None,
) -> svg.SVG:
    elements = list()

    # The maximum width we have reached for this picture
    canvas_width: float = 0
    # Default values needed for drawing
    if non_coding:
        total_width = min(sum(exon.size for exon in exons), max_width)
    else:
        total_width = min(sum(exon.coding.size for exon in exons), max_width)

    if height is None:
        height = min(0.1 * total_width, 20)

    if gap_size is None:
        exon_gap = min(height / 4, 5)
    else:
        exon_gap = gap_size

    # These values will be updated as we draw the figure
    x_position: float = 10
    y_position: float = 0

    for exon in exons:
        # The visual size of the exon depends on wether or not we draw the non-coding
        # regions
        if non_coding:
            exon_size = exon.size
        else:
            exon_size = exon.coding.size

        # If we overflow the width, go to a new line
        if x_position + exon_size + height > max_width / scale:
            # If we are still at the start position for x, we dont start a new line
            # (this happens when the exon we want to draw is larger than max_width)
            if x_position == 10:
                continue
            x_position = 10
            y_position += 2 * height

        for section in draw_exon(exon, height, reverse, non_coding):
            # Don't draw the empty sections
            if not section:
                continue
            # Shift the points
            section = shift(section, x_position, y_position)

            # Scale the points
            section = [x * scale for x in section]

            # fill = "green" if exon.frame == exon.end_frame else "black"
            fill = "#4C72B7"

            elements.append(
                svg.Polygon(
                    points=section, stroke="red", fill=fill, stroke_width=0  # type: ignore
                )
            )
        x_position = x_position + exon_size + exon_gap
        canvas_width = max(x_position, canvas_width)

    return svg.SVG(width=canvas_width, height=y_position + height, elements=elements)  # type: ignore


def draw_non_coding(region: Region, height: float) -> List[float]:
    """Draw a non coding region"""
    if not region:
        return list()

    top_left = (0, 0.25 * height)
    top_right = (region.size, 0.25 * height)
    bottom_right = (region.size, 0.75 * height)
    bottom_left = (0, 0.75 * height)

    # We need to end top-right, to draw the next thing
    return [*top_right, *bottom_right, *bottom_left, *top_left, *top_right]


def draw_coding(exon: Exon, height: float) -> List[float]:
    # Fixed points for every exon
    top_left = (0, 0)
    top_right = (exon.coding.size, 0)
    bottom_right = (exon.coding.size, height)
    bottom_left = (0, height)

    # Pre-calculate the ends for the 6 possible frames
    # Use the frame to fetch the index
    # fmt: off
    left_end = [
        tuple(), #  Straigt line, we don't need to draw anything
        (0.5*height, 0.5*height), #  point
        (-0.5*height, 0.5*height), #  notch
    ]

    right_end = [
        tuple(), #  straight line, we don't need to draw anything
        (exon.coding.size+0.5*height, 0.5*height), #  point
        (exon.coding.size-0.5*height, 0.5*height), #  notch

    ]
    # fmt: on

    return [
        *top_left,
        *top_right,
        *right_end[exon.end_frame],
        *bottom_right,
        *bottom_left,
        *left_end[exon.frame],
        *top_left,
    ]


Section = List[float]


def draw_exon(
    exon: Exon, height: float, reverse: bool, non_coding: bool
) -> Tuple[Section, Section, Section]:
    """Draw the specified exon by calculating all the points that have to be connected

    We draw clockwise, starting from the top left

    """

    # Get the non coding regions
    if non_coding:
        before_coding, after_coding = exon.non_coding
    else:
        before_coding = after_coding = Region(0, 0)

    x_offset: float

    if not reverse:
        before = draw_non_coding(before_coding, height)
        coding = shift(
            draw_coding(exon, height), x_offset=before_coding.size, y_offset=0
        )

        # Depending on the end of the coding part (notch or point), we have to shift the
        # non-coding part so it matches
        x_offset = before_coding.size + exon.coding.size
        if exon.end_frame == 1:
            x_offset += 0.5 * height
        elif exon.end_frame == 2:
            x_offset -= 0.5 * height

        after = shift(
            draw_non_coding(after_coding, height), x_offset=x_offset, y_offset=0
        )

        # We need to draw the coding region last, so it overlaps properly
    else:
        after = draw_non_coding(after_coding, height)
        coding = shift(
            draw_coding(exon, height), x_offset=after_coding.size, y_offset=0
        )

        # Depending on the end of the coding part (notch or point), we have to shift the
        # non-coding part so it matches
        x_offset = after_coding.size + exon.coding.size
        if exon.end_frame == 1:
            x_offset += 0.5 * height
        elif exon.end_frame == 2:
            x_offset -= 0.5 * height

        before = shift(
            draw_non_coding(before_coding, height), x_offset=x_offset, y_offset=0
        )

    if exon.coding:
        return (before, after, coding)
    else:
        return (before, after, list())



from typing import List

from exonviz.draw import draw_coding, draw_non_coding
from exonviz.draw import shift
from exonviz.exon import Exon, Region


# Describe the expected output for a given start and end frame
# fmt: off
targets = {
    "frame0_0": [
        0, 0,
        21, 0,
        21, 10,
        0, 10,
        0, 0
    ],
    "frame0_1": [
        0, 0,
        22, 0,
        17, 5,
        22, 10,
        0, 10,
        0, 0
    ]
}
# fmt: on


def test_in_frame() -> None:
    """ "This draws a square, like this: |||||"""
    exon = Exon(0, 21, frame=0, coding=Region(0, 21))
    height = 10

    # fmt: off
    target = [
        0, 0,
        21, 0,
        21, 10,
        0, 10,
        0, 0
    ]
    # fmt: on

    assert draw_coding(exon, height) == target


def test_end_frame_1() -> None:
    """Looks like this: ||||>"""
    exon = Exon(start=0, end=22, frame=0, coding=Region(0, 22))
    height = 10

    # fmt: off
    target = [
        0, 0,
        22, 0,
        27, 5,
        22, 10,
        0, 10,
        0, 0
    ]
    # fmt: on

    assert draw_coding(exon, height) == target


def test_no_shift() -> None:
    points: List[float] = [0, 0, 0, 0, 10]
    assert shift(points, 0, 0) == [0, 0, 0, 0, 10]


def test_shift_x_only() -> None:
    """Shift the X by 10"""
    points: List[float] = [0, 0, 0, 10]
    assert shift(points, 10, 0) == [10, 0, 10, 10]


def test_shift_x_y() -> None:
    """Shift both X and Y by 10"""
    points: List[float] = [0, 0, 0, 10]
    assert shift(points, 10, 8) == [10, 8, 10, 18]


def test_draw_non_coding() -> None:
    R = Region(0, 12)
    height = 10

    # fmt: off
    expected = [
        12, 2.5,
        12, 7.5,
        0, 7.5,
        0, 2.5,
        12, 2.5,
    ]
    # fmt: on
    assert draw_non_coding(R, height=height) == expected


def test_draw_non_coding_empty_region() -> None:
    R = Region(0, 0)
    height = 10

    assert draw_non_coding(R, height) == list()



from typing import Dict, Any, Optional, List, Tuple
import urllib.request
from urllib.error import HTTPError
import json

from .exon import Exon, Region


def mutalyzer(variant: str) -> Optional[Dict[str, Any]]:
    """Fetch variant information from mutalyzer"""

    url = f"https://mutalyzer.nl/api/normalize/{variant}"

    try:
        response = urllib.request.urlopen(url)
    except HTTPError as e:
        msg = f"Fetching '{url}' returned {e}"
        raise RuntimeError(msg)
    else:
        js = json.loads(response.read())
    selector: Dict[str, Any] = js["selector_short"]
    return selector


def convert_coding_positions(positions: List[List[str]]) -> Tuple[int, int]:
    start = int(positions[0][0])
    end = int(positions[0][1])

    if start > end:
        start, end = end, start

    start = start - 1
    return (start, end)


def is_reverse(positions: List[List[str]]) -> bool:
    """Is the transcript in reverse"""
    start = int(positions[0][0])
    end = int(positions[0][1])

    return start > end


def convert_exon_positions(positions: List[List[str]]) -> List[Tuple[int, int]]:
    converted = list()
    for mutalyzer_start, mutalyzer_end in positions:
        start = int(mutalyzer_start)
        end = int(mutalyzer_end)

        # Make sure start is always before end
        if start > end:
            start, end = end, start

        start = start - 1
        converted.append((start, end))
    return converted


def extract_exons(mutalyzer: Dict[str, Any]) -> Tuple[List[Exon], bool]:
    """Extract Exons from mutalyzer payload"""
    exons: List[Exon] = list()

    # Is the mutalyzer code in reverse
    reverse = is_reverse(mutalyzer["cds"]["g"])

    # Get the coding region
    coding = Region(*convert_coding_positions(mutalyzer["cds"]["g"]))

    # The first exon starts in frame by definition
    frame = 0

    for start, end in convert_exon_positions(mutalyzer["exon"]["g"]):
        E = Exon(start, end, frame, coding)
        frame = E.end_frame
        exons.append(E)

    return exons, reverse



from typing import List, Union, Tuple


class Region:
    """A region has a start and end, and an implicit size"""

    def __init__(self, start: int, end: int):
        """Ensures that start is always before end"""
        if end < start:
            raise ValueError("Reverse regions are not allowed (Region{start}, {end})")
        self.start = start
        self.end = end

    @property
    def size(self) -> int:
        """The size property."""
        return abs(self.end - self.start)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Region):
            return NotImplemented
        return self.start == other.start and self.end == other.end

    def __bool__(self) -> bool:
        return self.size != 0

    def __sub__(self, other: "Region") -> Tuple["Region", "Region"]:
        """Subtracting a region from itself returns the remaining two regions.

        Namely before and after the subtracted regio
        """
        # If other does not overlap and lies before
        if other.end < self.start:
            return Region(self.start, self.start), self

        # If other does not overlap and lies after
        if other.start > self.end:
            return self, Region(self.end, self.end)

        # Determine what is left before the subtracted regio
        start = self.start
        if other.start < self.start:
            end = self.start
        else:
            end = other.start
        before = Region(start, end)

        # Determine what is left after the subtracted regio
        if other.end > self.end:
            start = self.end
        else:
            start = other.end
        end = self.end
        after = Region(start, end)

        return (before, after)

    def __repr__(self) -> str:
        return f"Region(start={self.start}, end={self.end}, size={self.size})"


class Exon(Region):
    def __init__(
        self,
        start: int,
        end: int,
        frame: int,
        coding: Region,
    ) -> None:
        super().__init__(start, end)
        self.frame = frame
        self.coding = self._determine_coding(coding)

        self.non_coding = self._determine_non_coding()

    def __repr__(self) -> str:
        return (
            f"Exon(start={self.start}, end={self.end}, frame={self.frame}, "
            f"coding={self.coding}, non_coding={self.non_coding})"
        )

    def _determine_coding(self, coding: Region) -> Region:
        if coding.start > self.end:
            coding = Region(self.end, self.end)
        elif coding.end < self.start:
            coding = Region(self.start, self.start)
        else:
            coding = Region(max(self.start, coding.start), min(self.end, coding.end))
        return coding

    def _determine_non_coding(self) -> Tuple[Region, Region]:
        """Determine the non-coding parts of the exon"""
        exon_region = Region(self.start, self.end)
        return exon_region - self.coding

    @property
    def end_frame(self) -> int:
        return (self.coding.size + self.frame) % 3



import pytest
from exonviz.exon import Exon, Region


class TestRegion:
    def test_small_region(self) -> None:
        R = Region(0, 1)

        assert R.size == 1

    def test_empty_region(self) -> None:
        R = Region(10, 10)
        assert R.size == 0

    def test_empty_region_falsy(self) -> None:
        R = Region(10, 10)
        assert not R

    def test_non_empty_region_truthy(self) -> None:
        R = Region(0, 1)
        assert R

    def test_subtract_region_from_itself(self) -> None:
        """Subtracting a region from itself gives to emtpy Regions"""
        R = Region(0, 10)
        before, after = R - R

        assert not before
        assert not after

    def test_subtract_region_from_start(self) -> None:
        bigger = Region(0, 10)
        smaller = Region(0, 5)

        before, after = bigger - smaller

        assert not before
        assert after == Region(5, 10)

    def test_subtract_region_from_end(self) -> None:
        bigger = Region(0, 10)
        smaller = Region(5, 10)

        before, after = bigger - smaller

        assert before == Region(0, 5)
        assert not after

    def test_substract_region_from_center(self) -> None:
        bigger = Region(0, 10)
        smaller = Region(3, 7)

        before, after = bigger - smaller

        assert before == Region(0, 3)
        assert after == Region(7, 10)

    def test_subtract_larger_overlapping_start(self) -> None:
        bigger = Region(10, 20)
        smaller = Region(8, 13)

        before, after = bigger - smaller

        assert not before
        assert after == Region(13, 20)

    def test_subtract_larger_overlapping_end(self) -> None:
        bigger = Region(10, 20)
        smaller = Region(13, 25)

        before, after = bigger - smaller

        assert not after
        assert before == Region(10, 13)

    def test_subtract_larger_overlapping_all(self) -> None:
        bigger = Region(0, 20)
        smaller = Region(10, 15)

        before, after = smaller - bigger

        assert not before
        assert not after

    def test_subtract_unrelated_region_before(self) -> None:
        region = Region(0, 20)
        unrelated = Region(-20, -10)

        before, after = region - unrelated

        # This test relies on the implementation of 'before'
        assert before == Region(0, 0)
        assert after == Region(0, 20)

    def test_subtract_unrelated_region_after(self) -> None:
        region = Region(0, 20)
        unrelated = Region(40, 60)

        before, after = region - unrelated

        assert before == Region(0, 20)
        # This test relies on the implementation of 'after'
        assert after == Region(20, 20)

    def test_subtract_empty_region(self) -> None:
        region = Region(0, 20)
        empty = Region(10, 10)

        before, after = region - empty

        assert before == Region(0, 10)
        assert after == Region(10, 20)

    def test_reverse_region(self) -> None:
        """Test to make sure we reject reverse regions"""
        with pytest.raises(ValueError):
            region = Region(20, 10)


class TestExon:
    end_frame = [
        # Exon, expected end_frame
        # fmt: off
        (Exon(start=0, end=3, frame=0, coding=Region(0, 3)), 0),
        (Exon(start=0, end=4, frame=0, coding=Region(0, 4)), 1),
        (Exon(start=0, end=5, frame=0, coding=Region(0, 5)), 2),

        (Exon(start=0, end=3, frame=1, coding=Region(0, 3)), 1),
        (Exon(start=0, end=4, frame=1, coding=Region(0, 4)), 2),
        (Exon(start=0, end=5, frame=1, coding=Region(0, 5)), 0),

        (Exon(start=0, end=3, frame=2, coding=Region(0, 3)), 2),
        (Exon(start=0, end=4, frame=2, coding=Region(0, 4)), 0),
        (Exon(start=0, end=5, frame=2, coding=Region(0, 5)), 1)
        # fmt: on
    ]

    def test_zero_size_Exon(self) -> None:
        E = Exon(start=0, end=0, frame=0, coding=Region(0, 0))
        assert E.size == 0

    def test_larger_Exon(self) -> None:
        """Test Exon size"""
        E = Exon(start=0, end=21, frame=0, coding=Region(0, 0))
        assert E.size == 21

    @pytest.mark.parametrize("exon,expected", end_frame)
    def test_Exon_end_frame(self, exon: Exon, expected: int) -> None:
        """Test the Exon ending frame"""
        assert exon.end_frame == expected

    def test_Exon_coding_not_in_frame(self) -> None:
        """Test an exon where the size is in frame, but the coding size is not"""
        E = Exon(start=0, end=21, frame=0, coding=Region(8, 21))
        assert E.end_frame == 1

    def test_Exon_coding_in_frame(self) -> None:
        """Test an exon where the size is not in frame, but the coding size is"""
        E = Exon(start=0, end=22, frame=0, coding=Region(9, 21))
        assert E.end_frame == 0

    def test_Exon_non_coding(self) -> None:
        """If there is no coding, the whole exon should be none coding

        Note that Exon.non_coding is always a list of Regions
        """
        E = Exon(start=0, end=21, frame=0, coding=Region(0, 0))
        before_coding, after_coding = E.non_coding

        assert not before_coding
        assert not E.coding
        assert after_coding == Region(0, 21)

    def test_Exon_coding(self) -> None:
        """If the whole Exon is coding, non_coding should be empty"""
        E = Exon(start=0, end=21, frame=0, coding=Region(0, 21))
        before_coding, after_coding = E.non_coding

        assert not before_coding
        assert E.coding == Region(0, 21)
        assert not after_coding

    def test_Exon_start_non_coding(self) -> None:
        """Test the non_coding region if the start of the exon is non coding"""
        E = Exon(start=0, end=21, frame=0, coding=Region(4, 21))
        before_coding, after_coding = E.non_coding

        assert before_coding == Region(0, 4)
        assert E.coding == Region(4, 21)
        assert not after_coding

    def test_Exon_end_non_coding(self) -> None:
        E = Exon(start=0, end=21, frame=0, coding=Region(0, 18))
        before_coding, after_coding = E.non_coding

        assert not before_coding
        assert after_coding == Region(18, 21)

    def test_Exon_start_end_non_coding(self) -> None:
        """If only the center of the exon is coding"""
        E = Exon(start=0, end=21, frame=0, coding=Region(10, 18))
        before_coding, after_coding = E.non_coding

        assert before_coding == Region(0, 10)
        assert after_coding == Region(18, 21)

    CODING_SLICE = [
        # Coding region to pass to exon, sliced coding region
        (Region(200, 300), Region(121, 121)),  # coding region starts after exon
        (Region(0, 90), Region(100, 100)),  # coding region stops before exon
        (Region(0, 100), Region(100, 100)),  # coding region stops as the exon starts
        (Region(0, 101), Region(100, 101)),  # coding region overlaps exon by 1
        (Region(100, 200), Region(100, 121)),  # coding region extends after
        (Region(0, 200), Region(100, 121)),  # coding region fully contains exon
    ]

    @pytest.mark.parametrize("coding_region, coding_region_exon", CODING_SLICE)
    def test_exon_slice_coding(
        self, coding_region: Region, coding_region_exon: Region
    ) -> None:
        """If the coding region is larger than the exon, slice it to the exon size"""
        E = Exon(start=100, end=121, frame=0, coding=coding_region)

        assert E.coding == coding_region_exon



"""
Module that contains the command line app, so we can still import __main__
without executing side effects
"""

import argparse
import re
import sys
import math

from typing import List, Tuple
from .draw import draw_exons
from .exon import Exon
from .mutalyzer import mutalyzer, extract_exons


def check_input(transcript: str) -> str:
    """Check the input from the user, and rewrite when needed"""
    # If it looks like there is no variant
    if re.match(r"^\w+\.\d+$", transcript):
        return f"{transcript}:c.="

    # If it looks like the user forgot the version number, there isn't much we can do
    if re.match(r"^\w+$", transcript):
        msg = "Please specify the version of the transcript you are interested in: "
        raise RuntimeError(msg + transcript)

    return transcript


def fetch_exons(transcript: str) -> Tuple[List[Exon], bool]:
    """Make or fetch the requested exons"""

    # Does the transcript format make sense?
    transcript = check_input(transcript)

    payload = mutalyzer(transcript)
    if payload:
        return extract_exons(payload)
    else:
        return list(), False


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Description of command.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("transcript", help="Transcript (with version) to visualise")
    parser.add_argument(
        "--max-width", type=int, help="Maximum width of the figure", default=math.inf
    )
    parser.add_argument("--height", type=int, help="Exon height", default=None)
    parser.add_argument(
        "--non-coding",
        action="store_true",
        default=False,
        help="Show non coding regions",
    )
    parser.add_argument("--gap", type=int, help="Gap between the exons", default=None)
    args = parser.parse_args()

    # Try to talk to mutalyzer
    try:
        exons, reverse = fetch_exons(args.transcript)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        exit(1)

    for exon in exons:
        print(exon, file=sys.stderr)
    plot = draw_exons(
        exons,
        reverse,
        max_width=args.max_width,
        height=args.height,
        non_coding=args.non_coding,
        gap_size=args.gap,
    )
    print(plot)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
from pathlib import Path

from setuptools import find_packages
from setuptools import setup


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="exonviz",
    version="0.1.1",
    license="AGPL-3.0",
    description="Visualise exons and their reading frames",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Redmar van den Berg",
    author_email="RedmarvandenBerg@lumc.nl",
    url="https://github.com/redmar-van-den-berg/exonviz",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    package_data={
        "exonviz": ["py.typed"]
    },
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        "Topic :: Utilities",
    ],
    project_urls={
        "Changelog": "https://github.com/redmar-van-den-berg/exonviz/blob/main/CHANGELOG.md",
        "Issue Tracker": "https://github.com/redmar-van-den-berg/exonviz/issues",
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires=">=3.8",
    install_requires=[
        "svg.py",
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    setup_requires=[
        "pytest-runner",
    ],
    entry_points={
        "console_scripts": [
            "exonviz=exonviz.cli:main",
        ]
    },
)
